import sys
import argparse
import logging

from .mylog import configure_arg_parser_for_log, configure_logger
from .ws.ws_worker import websocket_run
from .soap.soap_worker import soap_run
from .test_conn import test_con
from .charger import ChargerInfo
from .constraints import CONSTRAINTS
from .config import CONFIG

def main():
    parser = argparse.ArgumentParser()
    configure_arg_parser_for_log(parser)
    parser.add_argument('--cloud', help='Cloud', default='default')
    parser.add_argument('--charger', help='Charger')
    parser.add_argument('--port', help='Server TCP/IP port', type=int)
    parser.add_argument('--server', help='Enable Server mode', action='store_true')
    parser.add_argument('--public', help='Public address of current station')
    parser.add_argument('--soap', help='Enable SOAP mode', action='store_true')
    parser.add_argument('--ws', help='Enable WebSocket mode', action='store_true')
    parser.add_argument('--web', help='Enable web interface', action='store_true')
    parser.add_argument('--test', help='Test connection', action='store_true')
    parser.add_argument('--id', help='Override charger id')
    env = parser.parse_args()

    # Logging
    configure_logger(env)
    log = logging.getLogger()

    # Arguments
    try:
        if env.soap and env.ws:
            raise UserWarning('You cannot combine WebSocket and SOAP modes')
            
    except UserWarning as ex:
        log.error(ex)
        return 1
        

    # Config
    config = CONFIG
    
    try:
        # Cloud
        clouds = config.get('clouds')
        if not clouds:
            raise UserWarning('Invalid config: element "clouds" not found')
            
        if env.cloud.startswith('ws:'):
            cloud_info = {
                'ws': env.cloud,
                'constraints':[]
            }
            env.soap = False
            log.info('Set WebSocket custom enpoint: %s', env.cloud)
        elif env.cloud.startswith('http:') or env.cloud.startswith('https:'):
            cloud_info = {
                'soap': env.cloud,
                'constraints':[]
            }
            env.soap = True
            log.info('Set SOAP custom enpoint: %s', env.cloud)
        else:
            cloud_info = clouds.get(env.cloud)
            if not cloud_info:
                raise UserWarning(f'Cloud "{env.cloud}" does not exist. Available: {list(clouds.keys())}')
            log.info('Selected cloud: %s', env.cloud)
        
        # if user didnt select explicitly
        if not env.soap and not env.ws:
            if cloud_info.get('ws'):  # if cloud supports websockets
                log.info('Implicitly selected WebSocket mode for cloud')
                env.soap = False
            else:
                log.info('Implicitly selected SOAP mode for cloud')
                env.soap = True
        
        # Check cloud constraints
        for cl_name, cl_data in clouds.items():
            if isinstance(cl_data, dict):    # skip "default": "foo"
                if 'constraints' in cl_data:
                    for c in cl_data['constraints']:
                        if not c in CONSTRAINTS:
                            raise UserWarning(f'Unknown constraint "{c}" in cloud "{cl_name}". Known constraints: {list(CONSTRAINTS.keys())}')
                else:
                    cl_data['constraints'] = []
                
                
        # Charger
        chargers = config.get('chargers')
        hardware = config.get('hardware')
        if not chargers:
            raise UserWarning('Invalid config: element "chargers" not found')
        if not hardware:
            raise UserWarning('Invalid config: element "hardware" not found')
        
        charger_helper = list(chargers.keys()) + list(hardware.keys())
        
        if env.charger:
            charger_info = chargers.get(env.charger)
            if not charger_info:
                hw = hardware.get(env.charger)
                if not hw:
                    raise UserWarning(f'Charger "{env.charger}" does not exist. Available: {charger_helper}')
                charger_info = {
                  "id": env.id or "foo",
                  "model": env.charger,
                  "meta": {}
                }
            log.info('Selected charger: %s', env.charger)
        else:
            charger_info = chargers.get('default')
            if not charger_info:
                raise UserWarning(f'There is not default charger specified. Use --charger with values {charger_helper}')
            log.info('Selected charger: %s', 'default')
                
        ch_model = hardware[charger_info['model']]
        
        # Merge meta from charger instance and model
        all_meta = charger_info['meta']
        for k, v in ch_model['meta'].items():
            if k not in all_meta:
                all_meta[k] = v
            
        final_id = env.id or charger_info['id']
        
        if env.soap:
            my_charger_info = ChargerInfo(final_id, cloud_info['soap'], env.public, ch_model['nbConnectors'], all_meta, cloud_info)
            return test_con(env, my_charger_info, log) if env.test else soap_run(env, my_charger_info, config, log)
        else:
            my_charger_info = ChargerInfo(final_id, cloud_info['ws'], None, ch_model['nbConnectors'], all_meta, cloud_info)
            return test_con(env, my_charger_info, log) if env.test else websocket_run(env, my_charger_info, config, log)
        
    except UserWarning as ex:
        log.error(ex)
        return 1

    return 0

sys.exit(main())




CONFIG = {
    "clouds":{
        "default": "production",
        
        "local":{
            "soap": "http://localhost:8080/steve/services/CentralSystemService",
            "ws": "ws://localhost:8080/steve/websocket/CentralSystemService"
        },
        "steve":{
            "soap": "http://steve.v4d.online:8080/steve/services/CentralSystemService",
            "ws": "ws://steve.v4d.online:8080/steve/websocket/CentralSystemService"
        },
        "staging":{
            "-soap": "http://staging.v4d.online:8080/steve/services/CentralSystemService",
            "ws": "ws://staging.v4d.online:5101/ocpp15",
            "constraints":["cn.zaria.timestamp"]
        },
        "test.v4d.ru":{
            "--soap (tomcat is off)": "http://test.v4d.ru/soap",
            "ws": "ws://test.v4d.ru/isws/ocpp15",
            "constraints":["cn.zaria.timestamp"]
        },
        "setec":{
            "soap": "http://is.v4d.online:8889/soap",
            "constraints":["cn.zaria.timestamp"]
        }
    },
    
    "hardware":{
        "evlink":{
            "nbConnectors": 1,
            "meta":{
                "chargePointVendor": "Schneider Electric",
                "chargePointModel": "EVlink Smart Wallbox",
                "firmwareVersion": "3.2.0.12",
                "chargePointSerialNumber": "3N170740564A1S1B7551700014",
                "chargeBoxSerialNumber": "EVB1A22P4RI3N1712406002002503A057"
            }
        },
        "setec":{
            "nbConnectors": 3,
            "meta":{
                "chargePointVendor": "5454",
                "chargePointModel": "35",
                "chargePointSerialNumber": "65",
                "chargeBoxSerialNumber": "75",
                "firmwareVersion": "65",
                "iccid": "78",
                "imsi": "78",
                "meterType": "7",
                "meterSerialNumber": "8"
            }
        }
    },
    
    "chargers":{
        "default":{
            "id": "FOO",
            "model": "evlink",
            "meta":{
            }
        },
        "foo":{
            "id": "423242",
            "model": "evlink",
            "meta":{
                "chargeBoxSerialNumber": "ololo"
            }
        },
        "setec":{
            "id": "423242",
            "model": "setec",
            "meta":{
            }
        }
    },
    
    "tags":{
    }
}




"""
    Содержит процедуры, обеспечивающие ОТСТУП от стандартов для проверки оборудования
    и систем, имеющих несоответствия
"""

CONSTRAINTS = {
    # Integration Server слал время в формате ISO, только кол-во цифр на мс зачем-то 7, а не 6
    'cn.zaria.timestamp': True
}




import uuid
import json
import asyncio
from urllib.parse import urlparse

from autobahn.asyncio.websocket import WebSocketClientFactory, WebSocketClientProtocol

from ..charger import Charger
from ..ocpp15.base import OcppMessage, OcppProtocolException, OcppError
from ..time_tools import Timeout


OCPP_J_CALL = 2        # Client-to-Server
OCPP_J_CALLRESULT = 3  # Server-to-Client
OCPP_J_CALLERROR = 4   # Server-to-Client

class MyClientProtocol(WebSocketClientProtocol):
    Charger = None
    Log = None
    Loop = None

    def onOpen(self):
        self.Log.info ('WebSocket connection open')
        
        self._last_request = None   # (id, action, timeout, callback)

        self.Loop.call_later(1.0, self.onTimer)

    def onMessage(self, payload, isBinary):
        assert not isBinary

        log = self.Log
        
        msg = None
        try:
            msg = json.loads(payload.decode('utf8'))
            
            log.debug('Message from CS: %s', msg)
            
            if msg[0] == OCPP_J_CALL:
                _, req_id, req_action, req_payload = msg

                log.info('CALL from CS: %s %s', req_action, req_payload)
                
                try:
                    resp_payload = self.Charger.on_request_from_cs(req_action, req_payload)
                    assert isinstance(resp_payload, dict), resp_payload
                    
                    log.info('Response to CS: %s', resp_payload)

                    ws_msg = [OCPP_J_CALLRESULT, req_id, resp_payload]
                
                except OcppError as ocpp_err:
                    log.error('Error processing request %s: %s', msg, str(ocpp_err))
                    ws_msg = [OCPP_J_CALLERROR, req_id, ocpp_err.ErrorCode.value, str(ocpp_err), {}]
                
                log.debug('Response to CS: %s', ws_msg)                
                self.sendMessage(json.dumps(ws_msg).encode('utf8'))
                
            elif msg[0] == OCPP_J_CALLRESULT:

                _, resp_id, resp_payload = msg

                log.info('CALLRESULT from CS: %s', resp_payload)

                if self._last_request:
                    req_id, req_action, _, cb = self._last_request
                    
                    if resp_id != req_id:
                        raise OcppProtocolException(f'Request/response mismatch (probably timeout)')

                    self._last_request = None
                    if cb:
                        cb(resp_payload)
                else:
                    raise OcppProtocolException(f'Unexpected response (probably timeout)')
                    
            elif msg[0] == OCPP_J_CALLERROR:
                log.error('CALLERROR from CS: %s', msg)
            else:
                raise OcppProtocolException(f'Message type "{msg[0]}" unknown')
        
        except Exception as ex:
            log.error(msg)
            raise
        
    def onTimer(self):
        self.Loop.call_later(1.0, self.onTimer)

        log = self.Log
        
        if self._last_request:
            _, action, timeout, cb = self._last_request
            if timeout.is_expired():
                log.error('Timeout on request to CS %s', action)
                if cb:
                    cb(None)
                self._last_request = None   
            else:
                return
        
        assert not self._last_request
        
        tpl = self.Charger.get_request_to_cs()
        if tpl:
            action, payload, cb = tpl
            id = str(uuid.uuid4())
            
            self._last_request = (id, action, Timeout(5.0), cb)
            
            request_json = [OCPP_J_CALL, id, action, payload]

            log.info('Request to CS: %s %s', action, payload)
            log.debug('Request to CS: %s', request_json)
            
            self.sendMessage(json.dumps(request_json).encode('utf8'))

    def onClose(self, wasClean, code, reason):
        self.Log.error ('WebSocket connection closed. Code=%s; Reason=%s', code, reason)
        self.Loop.stop()

'''    
def test_con(env, charger_info, log):
    import socket
    import select

    url = charger_info.CentralStationAddress

    log.info(f'Test connection to {url}...')

    url_items = urlparse(url)    
    host = url_items.hostname
    port = url_items.port or 80
    path = url_items.path + '/' + charger_info.Id

    s = None
    try:
        log.info(f'1. Open raw connection to {host}:{port}')
        s = socket.create_connection((host, port), timeout=5.0)

        log.info(f'2. Send HTTP request GET {path}')
        req_lines = [
            f'GET {path} HTTP/1.1',
            'User-Agent: tiger',
            f'Host: {host}:{port}',
            'Upgrade: WebSocket',
            'Connection: Upgrade',
            'Pragma: no-cache',
            'Cache-Control: no-cache',
            'Sec-WebSocket-Key: omf+YISYGBBTwPNsGSiWtg==',
            'Sec-WebSocket-Protocol: ocpp1.5',
            'Sec-WebSocket-Version: 13',
            '\r\n',
        ]
        s.sendall('\r\n'.join(req_lines).encode('utf-8'))
    
        fd_read, fd_write, fd_err = select.select([s], [], [], 5.0)
        if s in fd_read:
            data = s.recv(1024)
            if data:
                log.info(data)
            else:
                raise Exception('Socket closed while reading')
        else:
            raise Exception('Timeout reading')
    
    
    except Exception as ex:
        log.error(ex)

    finally:
        try:
            if s: s.close()
            s = None
        except:
            pass
'''

def websocket_run(env, charger_info, config, log):
    #test_con(env, charger_info, log)

    ws_url = charger_info.CentralStationAddress
    if ws_url.endswith('/'):
        ws_url += charger_info.Id
    else:
        ws_url += '/' + charger_info.Id
    
    ws_url_items = urlparse(ws_url)

    my_charger = Charger(charger_info)
    
    log.info(f'WebSocket endpoint: {ws_url}')
    factory = WebSocketClientFactory(url=ws_url, protocols=['ocpp1.5'])
    factory.protocol = MyClientProtocol

    loop = asyncio.get_event_loop()

    MyClientProtocol.Charger = my_charger
    MyClientProtocol.Log = log
    MyClientProtocol.Loop = loop
    
    coro = loop.create_connection(factory, ws_url_items.hostname, ws_url_items.port or 80)
    try:
        try:
            loop.run_until_complete(coro)
        except (TimeoutError, ConnectionRefusedError) as ex:
            log.error(ex)
            return 1
            
        loop.run_forever()
        return 0
    finally:
        loop.close()     
        


import logging
from enum import Enum
from collections import namedtuple

from .ocpp15.base import *
from .ocpp15.types import *
from .time_tools import Timeout
from .ocpp_payload_reader import MsgReader

class ChargerState(Enum):
    Offline = 0     # also when rejected from boot notification
    WaitBootNotificationResponse = 1
    Accepted = 2


ChargerInfo = namedtuple('ChargerInfo', 'Id CentralStationAddress ChargerPublicAddress NbConnectors Meta CloudInfo')


class Charger:
    def __init__(self, charger_info):
        self.Info = charger_info
        self._state = ChargerState.Offline
        self._timeout = Timeout(0, expired=True)    # Запрещает любые новые запросы
        self._log = logging.getLogger()
    
        # List of tuples (action, payload, cb)
        self._req_queue = []
        self._heartbeat_timeout = None
    
    def _pend_request(self, action, payload, cb):
        self._req_queue.append((action, payload, cb))

    # Returns Payload or throw OcppError
    def on_request_from_cs(self, action: str, payload: dict) -> dict:
        return self._process_payload(self._respond(action, payload))
            
    def _respond(self, action: str, payload: dict) -> dict:
        log = self._log
        
        reader = MsgReader(action, payload, self.Info, log)
        # Reset
        if action == 'Reset':
            return {'status': ResetStatus.Accepted}
        
        # ChangeConfiguration
        elif action == 'ChangeConfiguration':
            k = reader.str('key')
        
            if k=='HeartBeatInterval':
                v = reader.int('value')
                log.info(f'Set Heartbeat to {v} sec')
                self._heartbeat_timeout = Timeout(v)
            else:
                v = reader.str('value')
                log.info(f'Set {k}={v}')
            
            return {'status': ConfigurationStatus.Accepted}
        
        # ChangeAvailability
        elif action == 'ChangeAvailability':
            # The id of the connector for which availability needs to change. Id '0' (zero) is used if the availability of the charge point as a whole needs to change.
            connectorId = reader.int('connectorId')
            type = reader.enum('type', AvailabilityType)
            
            return {'status': AvailabilityStatus.Accepted}
        
        # RemoteStartTransaction
        elif action == 'RemoteStartTransaction':
            connectorId = reader.int('connectorId', -1)
            idTag = reader.str('idTag', MaxLen.IdToken)
            
            return self._process_payload({'status': RemoteStartStopStatus.Accepted})
        
        # RemoteStopTransaction
        elif action == 'RemoteStopTransaction':
            transactionId = reader.int('transactionId')
            
            return self._process_payload({'status': RemoteStartStopStatus.Accepted})
        
        # ReserveNow
        elif action == 'ReserveNow':
            # [integer] Mandatory. This contains the id of the connector to be reserved. A value of 0 means that the reservation is not for a specific connector.
            connectorId = reader.int('connectorId')
            # [dateTime] Mandatory. This contains the date and time when the reservation ends.
            expiryDate = reader.dtm('expiryDate')
            # [IdToken] Mandatory. The identifier for which the Charge Box has to reserve a connector.
            idTag = reader.str('idTag', MaxLen.IdToken)
            # [IdToken] Optional. The parent id-tag.
            parentIdTag = reader.str('parentIdTag', MaxLen.IdToken, 0)
            # [integer] Mandatory. Unique id for this reservation.
            reservationId = reader.int('reservationId')
            
            return {'status': ReservationStatus.Accepted}
        
        # CancelReservation
        elif action == 'CancelReservation':
            reservationId = reader.int('reservationId')
            
            return {'status': CancelReservationStatus.Accepted}
        else:
            raise OcppError(f'Action {action} is not supported', OcppErrorCode.NotSupported)
            
    # (Action, Payload, Callback|None) or None
    def get_request_to_cs(self) -> tuple:    
        log = self._log

        if not self._timeout.is_expired():
            return None
            
        s = self._state
        
        if s == ChargerState.Offline:
            self._state = ChargerState.WaitBootNotificationResponse
            return ('BootNotification', self._boot_notification_payload(), self.cb_BootNotification)

        elif s == ChargerState.WaitBootNotificationResponse:
            return None

        elif s == ChargerState.Accepted:
            if self._heartbeat_timeout.is_expired():
                self._heartbeat_timeout.start()
                return ('Heartbeat', {}, None)
                
            return self._req_queue.pop(0) if self._req_queue else None
            
        else:
            raise Exception(f'Unknown charger state: {self._state}')
        
    def cb_BootNotification(self, resp_payload: (dict, None)):
        assert resp_payload is None or isinstance(resp_payload, dict), resp_payload
        log = self._log
        
        #if self._state != ChargerState.WaitBootNotificationResponse:
        #    raise Exception(f'Received BootNotification response in {self._state} state')

        if resp_payload is None:
            self._offline('Failed boot notification')
            return

        reader = MsgReader(None, resp_payload, self.Info, log)

        status = reader.enum('status', RegistrationStatus)
            
        if status == RegistrationStatus.Accepted:
            currentTime = reader.dtm('currentTime')
            heartbeatInterval = reader.int('heartbeatInterval')
            
            log.info(f'Set Heartbeat to {heartbeatInterval} sec')
            self._heartbeat_timeout = Timeout(heartbeatInterval)
            
            self._state = ChargerState.Accepted
            for i in range(self.Info.NbConnectors):
                payload = {
                    'connectorId': i + 1,
                    'status': 'Available',
                    'errorCode': 'NoError',
                    'timestamp': utc_str(),
                }
                self._pend_request('StatusNotification', payload, None)

        else:
            self._offline(f'Charger rejected on cs: {resp_payload}. Device moved to Offline')
        
    def _boot_notification_payload(self) -> dict:
        msg_def = BootNotificationReq
        payload = {}
    
        for field_name, is_optional in msg_def:
            if field_name in self.Info.Meta:
                payload[field_name] = self.Info.Meta[field_name]
            elif not is_optional:
                raise Exception(f'Field "{field_name}" is required, but missing in charger {self.Info}')
        
        return payload
    
    def _offline(self, msg, timeout = 5.0):
        self._log.error(msg)
        self._state = ChargerState.Offline
        self._timeout.start(timeout)
        self._req_queue = []
    
    def manual_Authorize(self, tag: str):
        pld = {
            'idTag':    tag
        }
        self._pend_request('Authorize', pld, None)
    
    def _process_payload(self, payload: dict) -> dict:
        """
            Заменяет в иерархическом JSON объекте все Enum на их строковые значения
        """
        for k, v in payload.items():
            if isinstance(v, Enum):
                payload[k] = v.value
            elif isinstance(v, dict):
                process_payload(v)
        return payload


import datetime

from .ocpp15.types import OcppErrorCode
from .ocpp15.base import OcppError
from .constraints import CONSTRAINTS


class MsgReader:
    def __init__(self, action: (str, None), payload: dict, charger_info, log):
        assert action is None or isinstance(action, str), action
        assert isinstance(payload, dict), payload

        self._action = action
        self._payload = payload
        self._charger_info = charger_info
        self.log = log

    def str(self, name: str, max_len: int = None, def_val=None):
        assert name and isinstance(name, str), name

        if name not in self._payload:
            if def_val is None:
                raise OcppError(f'Missing "{name}" in payload {self._payload}', OcppErrorCode.PropertyConstraintViolation)
            return def_val
            
        v = self._payload[name]
        if max_len and len(v) > max_len:
            raise OcppError(f'Field "{name}={v}" in payload {self._payload} violates max length {max_len} (it is {len(v)} chars).', OcppErrorCode.PropertyConstraintViolation)

        return v

    def int(self, name: str, def_val=None):
        assert name and isinstance(name, str), name
        
        if name not in self._payload:
            if def_val is None:
                raise OcppError(f'Missing "{name}" in payload {self._payload}', OcppErrorCode.PropertyConstraintViolation)
            return def_val
            
        v = self._payload[name]
        try:
            return int(v)
        except ValueError:
            raise OcppError(f'Field "{name}={v}" in payload {self._payload} is invalid integer.', OcppErrorCode.PropertyConstraintViolation)

    def dtm(self, name: str, def_val=None):
        assert name and isinstance(name, str), name

        if name not in self._payload:
            if def_val is None:
                raise OcppError(f'Missing "{name}" in payload {self._payload}', OcppErrorCode.PropertyConstraintViolation)
            return def_val
            
        v = self._payload[name]
        try:
            return datetime.datetime.strptime(v, '%Y-%m-%dT%H:%M:%S.%fZ')    # %f is microsec, xxxxxx
            #return datetime.datetime.fromisoformat(v)  # from Python 3.7
        except ValueError as ex:
            pass
            
        # один лишний символ в мс
        v = v[:-2] + 'Z'
        try:
            res = datetime.datetime.strptime(v, '%Y-%m-%dT%H:%M:%S.%fZ')    # %f is microsec, xxxxxx
            if 'cn.zaria.timestamp' not in self._charger_info.CloudInfo['constraints']:
                self.log.warn('DateTime is malformed. Use "cn.zaria.timestamp" constraint to supress this warning')
            return res
        except ValueError as ex:
            raise OcppError(f'Field "{name}={v}" in payload {self._payload} is invalid DATE: {ex}', OcppErrorCode.PropertyConstraintViolation)
 

    def enum(self, name: str, cls=None, def_val=None):
        assert name and isinstance(name, str), name

        if name not in self._payload:
            if def_val is None:
                raise OcppError(f'Missing "{name}" in payload {self._payload}', OcppErrorCode.PropertyConstraintViolation)
            return def_val
            
        v = self._payload[name]
        try:
            return cls(v)
        except ValueError:
            raise OcppError(f'Field "{name}={v}" in payload {self._payload} is invalid. Possible values: {list(cls.__members__)}', OcppErrorCode.PropertyConstraintViolation)
            
            



from enum import Enum

"""
    Определяет максимальные размеры строковых типов
"""
class MaxLen:
    IdToken = 20
    SerialNumberString = 25


class ChargePointErrorCode(Enum):
    ConnectorLockFailure = 'ConnectorLockFailure'
    GroundFailure = 'GroundFailure'
    HighTemperature = 'HighTemperature'
    Mode3Error = 'Mode3Error'
    NoError = 'NoError'
    OtherError = 'OtherError'
    OverCurrentFailure = 'OverCurrentFailure'
    PowerMeterFailure = 'PowerMeterFailure'
    PowerSwitchFailure = 'PowerSwitchFailure'
    ReaderFailure = 'ReaderFailure'
    ResetFailure = 'ResetFailure'
    UnderVoltage = 'UnderVoltage'
    WeakSignal = 'WeakSignal'


class RemoteStartStopStatus(Enum):
    Accepted = 'Accepted'
    Rejected = 'Rejected'


class AvailabilityType(Enum):
    Inoperative = 'Inoperative' # Charge point is not available for charging.
    Operative = 'Operative'     # Charge point is available for charging.


class AvailabilityStatus(Enum):
    Accepted = 'Accepted'       # Request has been accepted and will be executed.
    Rejected = 'Rejected'       # Request has not been accepted and will not be executed.
    Scheduled = 'Scheduled'     # Request has been accepted and will be executed when transaction(s) in progress have finished.


class RegistrationStatus(Enum):
    Accepted = 'Accepted'   # Charge point is accepted by Central System.
    Rejected = 'Rejected'   # Charge point is not accepted by Central System. This happens when the charge point id is not known by Central System.


class ReservationStatus(Enum):
    Accepted = 'Accepted'   #Reservation has been made.
    Faulted = 'Faulted'     #Reservation has not been made, because connectors or specified connector are in a faulted state.
    Occupied = 'Occupied'   #Reservation has not been made. All connectors or the specified connector are occupied.    
    Rejected = 'Rejected'   #Reservation has not been made. Charge Box is not configured to accept reservations.
    Unavailable = 'Unavailable' # Reservation has not been made, because connectors or specified connector are in an unavailable state.


class CancelReservationStatus(Enum):
    Accepted = 'Accepted'   # Reservation for the identifier has been cancelled
    Rejected = 'Rejected'   # Reservation could not be cancelled, because there is no reservation active for the identifier.


class ResetStatus(Enum):
    Accepted = 'Accepted'
    Rejected = 'Rejected'


class ResetType(Enum):
    Hard = 'Hard'   # Full reboot of charge box software.
    Soft = 'Soft'   # Return to initial status, gracefully terminating any transactions in progress.


class ConfigurationStatus(Enum):
    Accepted = 'Accepted'
    Rejected = 'Rejected'           # Configuration key supported, but setting could not be changed.
    NotSupported = 'NotSupported'   # Configuration key is not supported.


class OcppErrorCode(Enum):
    # Requested Action is not known by receiver
    NotImplemented = 'NotImplemented'
    
    # Requested Action is recognized but not supported by the receiver
    NotSupported = 'NotSupported'
    
    # An internal error occurred and the receiver was not able to process the requested 
    # Action successfully
    InternalError = 'InternalError'
    
    # Payload for Action is incomplete
    ProtocolError = 'ProtocolError'
    
    # During the processing of Action a security issue occurred preventing receiver from completing the
    # Action successfully
    SecurityError = 'SecurityError'
    
    # Payload for Action is syntactically incorrect or not
    # conform the PDU structure for Action    
    FormationViolation = 'FormationViolation'
    
    # Payload is syntactically correct but at least one field contains an invalid value
    PropertyConstraintViolation = 'PropertyConstraintViolation'
    
    # Payload for Action is syntactically correct but at least one of the fields 
    # violates occurence constraints
    OccurenceConstraintViolation = 'OccurenceConstraintViolation'
    
    # Payload for Action is syntactically correct but at least one of the 
    # fields violates data type constraints (e.g. “somestring”: 12)
    TypeConstraintViolation = 'TypeConstraintViolation'
    
    # Any other error not covered by the previous ones
    GenericError = 'GenericError'



import sys
import argparse
import logging
import json

from .mylog import configure_arg_parser_for_log, configure_logger
from .client_websocket import websocket_run
from .client_soap import soap_run
from .charger import ChargerInfo

'''
class OcppServer(http.server.BaseHTTPRequestHandler):
    def do_POST(self):
        content_len = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_len)
        req = parse_req(body)
        
        print(req)
        
        if req.Action == 'BootNotification':
            res_payload = {
                'status': 'Accepted',
                'currentTime': _current_time_utc(),
                'heartbeatInterval': 5000
            }
            
        elif req.Action == 'StatusNotification':
            res_payload = {
            }          
            
        elif req.Action == 'Heartbeat':
            res_payload = {
                'currentTime': _current_time_utc(),
            }          
            
        elif req.Action == 'Authorize':
            res_payload = {
                'idTagInfo': {
                    'status': 'Accepted'
                }
            }          
            
        else:
            print ('Error', self.path, body)
            raise Exception (f'Unknown action: {req.Action}')
            
        res_bytes = build_res(req, req.Action + 'Response', res_payload)
        self.send_response(200)
        self.send_header('Content-type', 'application/soap+xml; charset=utf-8;')
        self.end_headers()
        self.wfile.write(res_bytes)
'''





def lcamel(s: str) -> str:
    assert isinstance(s, str), s
    return s[0].lower() + s[1:]
    
    
def ucamel(s: str) -> str:
    assert isinstance(s, str), s
    return s[0].upper() + s[1:]


class SoapException(UserWarning):
    def __init__(self, msg: str):
        super().__init__(msg)
    
FAULT_ACTION = 'urn://Ocpp/Cs/2012/06/:CentralSystemService:StatusNotification:Fault:SoapFault'
FAULT_PAYLOAD_NODE_NAME = 'Fault'
    
def is_fault_action(action:str) -> bool:
    return action == FAULT_ACTION


def fqid(id: str) -> str:
    return 'urn:uuid:' + id


def unqualify_id(id: str) -> str:
    return id[len('urn:uuid:'):] if id.startswith('urn:uuid:') else id


def create_fault_msg(code: str, subcode: str, reason: str):
    payload = {
        'Code':{
            'Value': code,
            'Subcode':{
                'Value': subcode
            }
        },
        'Reason':{
            'Text': reason
        }
    }
    return SoapMessage(FAULT_ACTION, payload)

class SoapMessage:
    def __init__(self, action: (str, None), payload: dict):
        assert action is None or isinstance(action, str), action
        assert isinstance(payload, dict), payload
        
        self.Action = action
        self.MessageID = None
        self.ChargeBoxIdentity = None
        self.RelatesTo = None
        self.ChargerPublicAddress = None
        self.CentralStationAddress = None
        self.Payload = payload
        
        
    def is_fault(self):
        return is_fault_action(self.Action)
        
    def fault_text(self) -> str:
        assert self.is_fault(), self
        assert self.Payload, self
        assert 'Reason' in self.Payload, self
        assert 'Text' in self.Payload['Reason'], self
        
        return self.Payload['Reason']['Text']
        
    def __str__(self):
        return f'SoapMessage<Action={self.Action} MessageID={self.MessageID} ChargeBoxIdentity={self.ChargeBoxIdentity} RelatesTo={self.RelatesTo} ChargerPublicAddress={self.ChargerPublicAddress} CentralStationAddress={self.CentralStationAddress} Payload={self.Payload}>'




import socket
import select
from urllib.parse import urlparse


def test_con(env, charger_info, log):
    url = charger_info.CentralStationAddress

    log.info(f'Test connection to {url}...')

    url_items = urlparse(url)    
    host = url_items.hostname
    port = url_items.port or 80
    path = url_items.path + '/' + charger_info.Id

    s = None
    try:
        log.info(f'1. Open raw connection to {host}:{port}')
        s = socket.create_connection((host, port), timeout=5.0)

        log.info(f'2. Send HTTP request GET {path}')
        req_lines = [
            f'GET {path} HTTP/1.1',
            'User-Agent: tiger',
            f'Host: {host}:{port}',
            'Upgrade: WebSocket',
            'Connection: Upgrade',
            'Pragma: no-cache',
            'Cache-Control: no-cache',
            'Sec-WebSocket-Key: omf+YISYGBBTwPNsGSiWtg==',
            'Sec-WebSocket-Protocol: ocpp1.5',
            'Sec-WebSocket-Version: 13',
            '\r\n',
        ]
        s.sendall('\r\n'.join(req_lines).encode('utf-8'))
    
        fd_read, fd_write, fd_err = select.select([s], [], [], 5.0)
        if s in fd_read:
            data = s.recv(1024)
            if data:
                log.info(data)
            else:
                raise Exception('Socket closed while reading')
        else:
            raise Exception('Timeout reading')
    
    
    except Exception as ex:
        log.error(ex)

    finally:
        try:
            if s: s.close()
            s = None
        except:
            pass



import xml.etree.ElementTree as ET

from .soap import SoapMessage, lcamel, ucamel, SoapException, is_fault_action, unqualify_id


def _local_name(name: str) -> str:
    """
    Removes namespace from node name
    """
    assert name and isinstance(name, str), name
    
    return name.split('}')[1] if '}' in name else name


def _get_xml_node(root, path: str):
    """
        Throws SoapException
        
        path: a/b/c
    """
    assert path and isinstance(path, str), path

    res = root
    for short_name in path.split('/'):
        found = False
        for ch_node in res:
            if _local_name(ch_node.tag) == short_name:
                found = True
                res = ch_node
                break
        if not found:
            raise SoapException(f'Path "{path}" could not be resolved, stopped on "{short_name}"')
    return res


def _parse_payload(payload_node) -> dict:
    res = {}
    for nd in payload_node:
        nd_name = _local_name(nd.tag)
        if list(nd):  # has subnodes?
            res[nd_name] = _parse_payload(nd)
        else:
            res[nd_name] = nd.text
            
    return res
    
    
def _t(root, path: str) -> str:
    return _get_xml_node(root, path).text
        
    
def parse_message(xml_data: bytes, is_request: bool) -> SoapMessage: 
    try:
        root = ET.fromstring(xml_data)
    except Exception as ex:
        raise SoapException(f'Malformed XML data: {xml_data}')
    

    action = _t(root, 'Header/Action')

    if is_request:
        if not action.startswith('/'):
            raise SoapException(f'Action should start with "/": {action}')

        action = action[1:]
        
        payload_nd = _get_xml_node(root, 'Body/' + lcamel(action) + 'Request')
    else:
        if is_fault_action(action):
            payload_nd = _get_xml_node(root, 'Body/Fault')
        else:
            if not action.startswith('/'):
                raise SoapException(f'Action should start with "/": {action}')

            action = action[1:]

            #if not action.endswith('Response'):
            #    raise SoapException(f'Response Action should end with "Response": {action}')
        
            body_nd = _get_xml_node(root, 'Body')
            payload_nd = body_nd[0]
    
    res = SoapMessage(action, _parse_payload(payload_nd))
    res.MessageID = unqualify_id(_t(root, 'Header/MessageID'))
    
    if is_request:
        res.ChargeBoxIdentity = _t(root, 'Header/chargeBoxIdentity')
    else:
        res.RelatesTo = unqualify_id(_t(root, 'Header/RelatesTo'))
    
    return res




import uuid
import xml.etree.ElementTree as ET

from .soap import SoapMessage, lcamel, ucamel, fqid


def _create_text_elt(parent, name, text):
    elt = ET.SubElement(parent, name)
    elt.text = text
    return elt


def _build_payload(xml_nd, payload: dict):
    for k, v in payload.items():
        sub_elt = ET.SubElement(xml_nd, k)
        if isinstance(v, dict):
            _build_payload(sub_elt, v)
        else:
            sub_elt.text = str(v)


def build_msg(msg: SoapMessage, is_request: bool, is_cs: bool = False) -> bytes:
    """
        MessageID можно не заполнять - будет установлен автоматически
    """
    assert isinstance(msg, SoapMessage), msg
    assert msg.Action and isinstance(msg.Action, str), msg
    
    tree = ET.ElementTree()

    Envelope = ET.Element('soap:Envelope', attrib={
        'xmlns:soap':       'http://www.w3.org/2003/05/soap-envelope',
        #'xmlns:cp':         'urn://Ocpp/Cp/2012/06/',
        'xmlns:wsa5':       'http://www.w3.org/2005/08/addressing',
        'xmlns:cs':         'urn://Ocpp/Cs/2012/06/'
    })
    
    Header = ET.SubElement(Envelope, 'soap:Header')
    
    # chargeBoxIdentity
    if is_request:
        _create_text_elt(Header, 'cs:chargeBoxIdentity', msg.ChargeBoxIdentity)
    
    # Action
    Header_Action = ET.SubElement(Header, 'wsa5:Action', attrib={'soap:mustUnderstand': 'true'} if is_request else {})
    Header_Action.text = '/' + msg.Action
    
    # MessageID
    if not msg.MessageID:
        msg.MessageID = str(uuid.uuid4())
    _create_text_elt(Header, 'wsa5:MessageID', fqid(msg.MessageID))
    
    # To    
    Header_To = ET.SubElement(Header, 'wsa5:To', attrib={'soap:mustUnderstand': 'true'} if is_request else {})
    Header_To.text = msg.CentralStationAddress if is_request else 'http://www.w3.org/2005/08/addressing/anonymous'
    
    # From
    if is_request:
        if msg.ChargerPublicAddress:
            Header_From = ET.SubElement(Header, 'wsa5:From')
            _create_text_elt(Header_From, 'wsa5:Address', msg.ChargerPublicAddress)
        else:
            pass
    
    # ReplyTo
    if is_request:
        Header_ReplyTo = ET.SubElement(Header, 'wsa5:ReplyTo')
        _create_text_elt(Header_ReplyTo, 'wsa5:Address', 'http://www.w3.org/2005/08/addressing/anonymous')
    
    # RelatesTo
    if not is_request:
        Header_RelatesTo = ET.SubElement(Header, 'wsa5:RelatesTo')
        Header_RelatesTo.text = fqid(msg.RelatesTo)
    
    Body = ET.SubElement(Envelope, 'soap:Body')
    
    if msg.is_fault():
        payload_nd = ET.SubElement(Body, 'Fault', attrib={'xmlns': 'http://www.w3.org/2003/05/soap-envelope'})
    else:
        payload_nd_name = lcamel(msg.Action) 
        if is_request:
            payload_nd_name += 'Request'
        payload_ns = 'urn://Ocpp/Cs/2012/06/' if is_request else 'urn://Ocpp/Cp/2012/06/'
        payload_nd = ET.SubElement(Body, payload_nd_name, attrib={'xmlns': payload_ns})
        
    _build_payload(payload_nd, msg.Payload)
    
    return b'<?xml version="1.0" encoding="UTF-8"?>' + ET.tostring(Envelope)



import requests
import json
import time

import http.server
import socketserver
from urllib.parse import urlparse, parse_qs

from ..ocpp15.base import OcppMessage, OcppProtocolException, OcppError
from ..charger import Charger
from .soap import SoapMessage, create_fault_msg
from .soap_build import build_msg
from .soap_parse import parse_message

MIME={}
MIME['txt'] = 'text/plain; charset=utf-8'
MIME['html'] = 'text/html; charset=utf-8'
MIME['css'] = 'text/css'
MIME['js'] = 'application/javascript'
MIME['jpg'] = 'image/jpg'
MIME['json'] = 'application/json'
MIME['xml'] = 'application/xml; charset=utf-8'


def soap_request(req: SoapMessage, log) -> SoapMessage:
    assert isinstance(req, SoapMessage), req
    assert req.Action, req
    assert isinstance(req.Payload, dict), req   # empty {} is False
    
    log.info('Request to CS: %s %s', req.Action, req.Payload)
    
    #log.debug('Request to CS: %s', req)

    req_bytes = build_msg(req, is_request=True)

    log.debug('CP ==(req)=> CS: %s', req_bytes)
    
    # raises
    resp_http = requests.post(req.CentralStationAddress, data=req_bytes)
    
    log.debug('Response %s %s %s', resp_http.status_code, resp_http.headers, resp_http.content)
    
    resp = parse_message(resp_http.content, is_request=False)
        
    log.info('Response from CS: %s %s', resp.Action, resp.Payload)
    log.debug('Response from CS: %s', resp)
            
    # Check for faults
    if resp.is_fault():
        raise Exception(f'Fault for request {req}: {resp.fault_text()}')

    #expected_resp_action = req.Action + 'Response'
    #if resp.Action != expected_resp_action:
    #    raise Exception(f'Request/response mismatch. Expected response action {expected_resp_action} but got {resp.Action}. Response: {resp}')
        
    if resp.RelatesTo != req.MessageID:
        raise Exception(f'Request/response mismatch. Expected response RelatesTo={req.MessageID} but got  {resp.RelatesTo}. Response: {resp}')
        
    return resp


class WebInterfaceHttpServer(http.server.BaseHTTPRequestHandler):
    Charger = None
    Config = None
    Log = None
    
    def do_GET(self):
        if self.path == '/':
            self.path = '/index.html'
        
        url = urlparse(self.path)
        q = parse_qs(url.query)
        
        if url.path == '/control/authorize':
            self.Charger.manual_Authorize(q['idTag'][0])
            self.send_response(204)
            self.end_headers()
        elif url.path == '/monitor/tags':
            self.send_response(200)
            self.send_header('Content-type', MIME['json'])
            self.end_headers()
            self.wfile.write(json.dumps(self.Config['tags']).encode())
        else:
            dot = url.path.rfind('.')
            if dot != -1:
                ext = url.path[dot+1:]
                self.sendFile('./www/' + url.path, MIME.get(ext, MIME['txt']))
            else:
                self.sendFile('./www/' + url.path, MIME['txt'])
   
    def sendFile(self, path, mime):
        try:
            with open(path, 'rb') as f:
                self.send_response(200)
                self.send_header('Content-type', mime)
                self.end_headers()
                self.wfile.write(f.read())
        except IOError as ex:
            self.send_error(404, f'Error reading {path}: {ex}')
            
            
class SoapChargerServer(http.server.BaseHTTPRequestHandler):
    Charger = None
    Log = None

    '''
    def do_GET(self):
        if self.path == '/':
            self.path = '/index.html'
        
        url = urlparse(self.path)
        print (url.path)
        
        if url.path == '/control/authorize':
            self.Charger.postpone_authorize('17171771771')
            self.send_response(204)
            self.end_headers()
        else:
            dot = url.path.rfind('.')
            if dot != -1:
                ext = url.path[dot+1:]
                self.sendFile('./www/' + url.path, MIME.get(ext, MIME['txt']))
            else:
                self.sendFile('./www/' + url.path, MIME['txt'])
    '''    
    def do_POST(self):
        log = self.Log
        
        if self.path != '/':
            raise Exception(f'POST is only accepted on path "/", not "{self.path}"')
        
        content_len = int(self.headers.get('Content-Length', 0))
        body: bytes = self.rfile.read(content_len)
        req: SoapMessage = parse_message(body, is_request=True)

        log.info('Request from CS: %s %s', req.Action, req.Payload)
        log.debug('Request from CS: %s', req)

        try:
            resp_payload = self.Charger.on_request_from_cs(req.Action, req.Payload)
            
            resp = SoapMessage(req.Action + 'Response', resp_payload)
        except OcppError as ocpp_err:
            log.error('Error processing request %s: %s', req, str(ocpp_err))
            resp = create_fault_msg(ocpp_err.ErrorCode.value, 'SecurityError', str(ocpp_err))
        
        resp.ChargeBoxIdentity = self.Charger.Info.Id
        resp.RelatesTo = req.MessageID
        resp.ChargerPublicAddress = ''
        resp.CentralStationAddress = self.Charger.Info.CentralStationAddress

        log.info('Response to CS: %s %s', resp.Action, resp.Payload)
        
        resp_xml = build_msg(resp, is_request=False)
        log.debug('CP ==(resp)=> CS: %s', resp_xml)

        self.send_response(200)
        self.send_header('Content-type', MIME['xml'])
        self.end_headers()
        self.wfile.write(resp_xml)


class BothServers(WebInterfaceHttpServer, SoapChargerServer):
    pass

def run_charger_once(charger, log):
    """
        Обрабатывает выходные запросы от ЗС
    """
    tpl = charger.get_request_to_cs()
    if not tpl:
        return
        
    action, payload, cb = tpl
    
    msg = SoapMessage(action, payload)
    msg.ChargeBoxIdentity = charger.Info.Id
    msg.ChargerPublicAddress = charger.Info.ChargerPublicAddress
    msg.CentralStationAddress = charger.Info.CentralStationAddress

    resp = soap_request(msg, log)
    
    if cb:
        cb(resp.Payload)
    else:
        log.debug('Response ignored - no callback specified')

def soap_run(env, charger_info, config, log):
    if not charger_info.ChargerPublicAddress:
        log.warn ('Public address not specified, SOAP reverse commands wont work. Use --public "http://1.2.3.4:5678" argument')

    my_charger = Charger(charger_info)

    SoapChargerServer.Charger = my_charger
    SoapChargerServer.Log = log
    
    WebInterfaceHttpServer.Charger = my_charger
    WebInterfaceHttpServer.Config = config
    WebInterfaceHttpServer.Log = log
    
    log.info(f'SOAP Endpoint is {charger_info.CentralStationAddress}')
    
    # Сервер запускается если указан обратный адрес подключения к ЗС или нужен веб-интерфейс для управления
    if charger_info.ChargerPublicAddress or env.web:
        if env.port:
            port = env.port
            log.info(f'Listening on explicitly specified port {port}')
        else:
            # Если порт не указан явно, попытаемся получить его из публичного адреса
            port = None
            if charger_info.ChargerPublicAddress:
                url = urlparse(charger_info.ChargerPublicAddress)
                if url.port:
                    port = url.port
                    log.info(f'Listening on port {port} from public url {charger_info.ChargerPublicAddress}')
            
            if not port:
                port = 80
                log.info(f'Listening on default port {port}. Use --port option to change it')
            
        server_type = None
        if charger_info.ChargerPublicAddress and env.web:
            server_type = BothServers
        elif charger_info.ChargerPublicAddress:
            server_type = SoapChargerServer
        else:
            server_type = WebInterfaceHttpServer
            
        with socketserver.TCPServer(('', port), server_type) as httpd:
            #httpd.serve_forever()
            httpd.timeout = 0.5
            while True:
                httpd.handle_request()
                run_charger_once(my_charger, log)
    else:
        log.info('No HTTP server is setup - charger has no public address and web interface was not requested')
        while True:
            run_charger_once(my_charger, log)
            time.sleep(0.2)
    
    return 0


import datetime


Optional = True
Mandatory = False


BootNotificationReq = (
    ('chargeBoxSerialNumber',   Optional),
    ('chargePointModel',        Mandatory),
    ('chargePointSerialNumber', Optional),
    ('chargePointVendor',       Mandatory),
    ('firmwareVersion',         Optional),
    ('iccid',                   Optional),
    ('imsi',                    Optional),
    ('meterSerialNumber',       Optional),
    ('meterType',               Optional),
)


def utc_str() -> str:    
    return str(datetime.datetime.utcnow().isoformat(timespec='milliseconds')) + 'Z' # '2019-03-18T10:21:03.944Z'


class OcppMessage:
    def __init__(self, action: (str, None), payload: dict):
        assert isinstance(payload, dict), payload
        
        self.Action = action    # can be None
        self.MessageID = None
        self.Payload = payload
        
    def __str__(self):
        return f'OcppMessage<Action={self.Action} MessageID={self.MessageID} Payload={self.Payload}>'


class OcppException(UserWarning):
    def __init__(self, msg: str):
        super().__init__(msg)


class OcppProtocolException(OcppException):
    def __init__(self, msg: str):
        super().__init__(msg)


class OcppError(UserWarning):
    def __init__(self, msg: str, code):
        super().__init__(msg)
        self.ErrorCode = code

